#include "time/RosenbrockAdaptInstationary.h"
#include "AdaptInfo.h"
#include "ProblemTimeInterface.h"

namespace AMDiS {

  RosenbrockAdaptInstationary::RosenbrockAdaptInstationary(std::string name, 
							   RosenbrockStationary *problemStat,
							   AdaptInfo *info,
							   ProblemTimeInterface *problemInstat,
							   AdaptInfo *initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      fixFirstTimesteps(0),
      tau(1.0),
      tauGamma(1.0)
  {
    FUNCNAME("RosenbrockAdaptInstationary::RosenbrockAdaptInstationary()");

    std::string str;
    GET_PARAMETER(0, name + "->rosenbrock method", &str);
    RosenbrockMethodCreator *creator = 
      dynamic_cast<RosenbrockMethodCreator*>(CreatorMap<RosenbrockMethod>::getCreator(str));
    rosenbrockMethod = creator->create();

    TEST_EXIT_DBG(rosenbrockMethod)("This should not happen!\n");

    GET_PARAMETER(0, name + "->fix first timesteps", "%d", &fixFirstTimesteps);
    problemStat->setRosenbrockMethod(rosenbrockMethod);

    adaptInfo->setRosenbrockMode(true);

    problemStat->setTauGamma(&tauGamma);
    problemStat->setTau(&tau);
  }


  void RosenbrockAdaptInstationary::oneTimestep()
  {
    FUNCNAME("RosenbrockAdaptInstationary::oneTimestep()");

    // estimate before first adaption
    if (adaptInfo->getTime() <= adaptInfo->getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);

    bool rejected = false;
    double timeTol = adaptInfo->getTimeTolerance(0);
    
    do {
      // increment time
      adaptInfo->setTime(adaptInfo->getTime() + adaptInfo->getTimestep());
      problemTime->setTime(adaptInfo);
      tau = adaptInfo->getTimestep();
      tauGamma = 1.0 / (tau * rosenbrockMethod->getGamma());
      
      INFO(info, 6)("time = %e   timestep = %e\n",
		    adaptInfo->getTime(), adaptInfo->getTimestep());

      adaptInfo->setSpaceIteration(0);
      
      // do the iteration
      problemIteration->beginIteration(adaptInfo);
      problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
      problemIteration->endIteration(adaptInfo);

      double errorEst = adaptInfo->getTimeEstSum(0);
      double newTimestep = 0.0;
      double order = rosenbrockMethod->getOrder();

      if (errorEst < timeTol || fixFirstTimesteps > 0 || firstTimestep) {
	double fac = 1.0;
	
	if (fixFirstTimesteps > 0) {
	  newTimestep = adaptInfo->getTimestep();
	} else {
	  if (firstTimestep || succRejection) {
	    fac = pow((timeTol / errorEst), 1.0 / order);
	  } else {
	    fac = adaptInfo->getTimestep() / tauAcc * 
	      pow((timeTol * estAcc / (errorEst * errorEst)), 1.0 / order);
	  }
	  fac = std::min(fac, 3.0);
	  newTimestep = fac * adaptInfo->getTimestep();
	  newTimestep *= 0.95;
	}

	tauAcc = adaptInfo->getTimestep();
	estAcc = errorEst;
	
	lastTimestepRejected = false;
	succRejection = false;
      } else {
	if (lastTimestepRejected) {
	  succRejection = true;  
	  
	  double reducedP = 
	    log(errorEst / estRej) / log(adaptInfo->getTimestep() / tauRej);
	  
	  if (reducedP < order && reducedP > 0.0)
	    order = reducedP;
	} 
	
	newTimestep = pow((timeTol / errorEst), 1.0 / order) * adaptInfo->getTimestep();
	newTimestep *= 0.95;

	tauRej = adaptInfo->getTimestep();
	estRej = errorEst;
	
	lastTimestepRejected = true;
      }
              

      if (errorEst < timeTol || fixFirstTimesteps) {
	INFO(info, 6)("Accepted timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());

	adaptInfo->setTimestep(newTimestep);

	rejected = false;

	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));
      } else {
	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));

	INFO(info, 6)("Rejected timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());
	
	adaptInfo->setTime(adaptInfo->getTime() - adaptInfo->getTimestep());
	adaptInfo->setTimestep(newTimestep);

	rejected = true;
      }

      if (firstTimestep)
	firstTimestep = false;

      if (fixFirstTimesteps > 0)
	fixFirstTimesteps--;

    } while (rejected);

    rosenbrockStat->acceptTimestep();

    adaptInfo->setLastProcessedTimestep(adaptInfo->getTimestep());
    adaptInfo->incTimestepNumber();
  }

}
